#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/termios.h>
#include <sys/mman.h>
#include "main.h"
#include "ops.h"
#include "traps.h"

u16 memory[MEMORY_MAX];
u16 reg[R_COUNT];
u16 running;

// Helper functions
u16 sign_extend(u16 x, int bit_count) {
  if ((x >> (bit_count - 1)) & 1) {
    x |= (0xFFFF << bit_count);
  }
  return x;
}

void update_flags(u16 r) {
  if (reg[r] == 0) {
    reg[R_COND] = FL_ZRO;
  } else if (reg[r] >> 15) {
    reg[R_COND] = FL_NEG;
  } else {
    reg[R_COND] = FL_POS;
  }
}

u16 swap16(u16 x) {
  return (x << 8) | (x >> 8);
}

void read_image_file(FILE *file) {
  u16 origin;
  fread(&origin, sizeof(origin), 1, file);
  origin = swap16(origin);

  u16 max_read = MEMORY_MAX - origin;
  u16 *p = memory + origin;
  size_t read = fread(p, sizeof(u16), max_read, file);

  while (read-- > 0) {
    *p = swap16(*p);
    ++p;
  }
}

int read_image(const char *image_path) {
  FILE *file = fopen(image_path, "rb");
  if (!file) return 0;
  read_image_file(file);
  fclose(file);
  return 1;
}

// Terminal input stuff
struct termios original_tio;

void disable_input_buffering() {
  tcgetattr(STDIN_FILENO, &original_tio);
  struct termios new_tio = original_tio;
  new_tio.c_lflag &= ~ICANON & ~ECHO;
  tcsetattr(STDIN_FILENO, TCSANOW, &new_tio);
}

void restore_input_buffering() {
  tcsetattr(STDIN_FILENO, TCSANOW, &original_tio);
}

u16 check_key() {
  fd_set readfds;
  FD_ZERO(&readfds);
  FD_SET(STDIN_FILENO, &readfds);

  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 0;
  return select(1, &readfds, NULL, NULL, &timeout) != 0;
}

void handle_interrupt(int signal) {
  restore_input_buffering();
  printf("\n");
  exit(-2);
}

void mem_write(u16 address, u16 value) {
  memory[address] = value;
}

u16 mem_read(u16 address) {
  if (address == MR_KBSR) {
    if (check_key()) {
      memory[MR_KBSR] = (1 << 15);
      memory[MR_KBDR] = getchar();
    }
    else {
      memory[MR_KBSR] = 0;
    }
  }
  return memory[address];

}

int main(int argc, char **argv) {
  u16 instr, op;

  if (argc < 2) {
    printf("usage: lc3 filename [filename2 filename3 ...]\n");
    exit(2);
  }

  for (int i = 1; i < argc; i++) {
    if (!read_image(argv[i])) {
      printf("unable to parse %s\n", argv[i]);
      exit(1);
    }
  }

  signal(SIGINT, handle_interrupt);
  disable_input_buffering();

  // set condition flag
  reg[R_COND] = FL_ZRO;

  // set starting position of program counter
  reg[R_PC] = PC_START;

  running = 1;
  while (running) {
    instr = mem_read(reg[R_PC]++);
    // opcode is only the first 4 bits of instruction.
    op = instr >> 12;

    switch (op) {
      case OP_ADD:
        op_add(instr);
        break;
      case OP_AND:
        op_and(instr);
        break;
      case OP_NOT:
        op_not(instr);
        break;
      case OP_BR:
        op_br(instr);
        break;
      case OP_JMP:
        op_jmp(instr);
        break;
      case OP_JSR:
        op_jsr(instr);
        break;
      case OP_LD:
        op_ld(instr);
        break;
      case OP_LDI:
        op_ldi(instr);
        break;
      case OP_LDR:
        op_ldr(instr);
        break;
      case OP_LEA:
        op_lea(instr);
        break;
      case OP_ST:
        op_st(instr);
        break;
      case OP_STI:
        op_sti(instr);
        break;
      case OP_STR:
        op_str(instr);
        break;
      case OP_TRAP:
        reg[R_R7] = reg[R_PC];
        switch (instr & 0xFF)
        {
          case TRAP_GETC:
            trap_getc();
            break;
          case TRAP_OUT:
            trap_out();
            break;
          case TRAP_PUTS:
            trap_puts();
            break;
          case TRAP_IN:
            trap_in();
            break;
          case TRAP_PUTSP:
            trap_putsp();
            break;
          case TRAP_HALT:
            trap_halt();
            break;
        }
        break;
      case OP_RES:
      case OP_RTI:
      default:
        abort();
        break;

    }
  }

  restore_input_buffering();
}
